mod omori_locator;
mod app_logic;
mod config;

use std::{sync::Arc, thread, time::{Duration, SystemTime}};

use app_logic::{AppThread, UiEvent, UiState, UiStateHolder};
use crossbeam_channel::{Receiver, Sender};
use eframe::egui;
use egui::{mutex::{Mutex, RwLock}, Align, Layout, RichText, ThemePreference};
use egui_alignments::{center_horizontal, center_vertical, top_horizontal, Aligner};
use egui_extras::{Column, TableBuilder};
use egui_modal::Modal;
use sha2::Digest;

fn main() -> anyhow::Result<()> {
    env_logger::init();

    #[cfg(target_os = "windows")]
    log::trace!("How do you even do serious work on ValorantOS??");

    #[cfg(target_os = "macos")]
    log::trace!("Mac user, rich fuck, send me $500");
    
    #[cfg(target_os = "linux")]
    log::trace!("Linux user... Please touch grass for once");
    

    let options = eframe::NativeOptions {
        viewport: egui::ViewportBuilder::default()
            .with_min_inner_size([640.0, 480.0]),
        ..Default::default()
    };

    let app_state = UiStateHolder {
        state: Arc::new(Mutex::new(UiState::Loading))
    };

    let (sender, receiver): (Sender<UiEvent>, Receiver<UiEvent>) = crossbeam_channel::unbounded();

    eframe::run_native("TundleBool", options, Box::new(|cc| {
        cc.egui_ctx.set_theme(ThemePreference::System);
        cc.egui_ctx.style_mut(|style| {
            style.interaction.selectable_labels = false;
        });
        
        let mut app_thread = AppThread::create(&app_state, &cc.egui_ctx, &receiver);

        thread::spawn(move || {
            app_thread.main();
        });

        Ok(Box::<Application>::new(Application::create(sender, app_state)))
    })).unwrap();

    Ok(())
}


struct Application {
    state: UiStateHolder,
    sender: Sender<UiEvent>,
    key_input: String
}

impl Application {
    fn create(sender: Sender<UiEvent>, state: UiStateHolder) -> Application {
        Application {
            sender,
            state,
            key_input: "".to_string()
        }
    }
}

const GAME_KEY_HASH: &[u8; 32] = include_bytes!("keyhash");

impl eframe::App for Application {
    fn update(&mut self, ctx: &egui::Context, frame: &mut eframe::Frame) {
        let state = self.state.state.lock();
        let state = state.clone();

        match state {
            UiState::Loading => {
                egui::CentralPanel::default().show(ctx, |ui| {
                    center_vertical(ui,|ui| { 
                        ui.label(RichText::new("Loading").size(32.0));
                    });
                });
            },
            UiState::Error(reason) => {
                egui::TopBottomPanel::top("error_title_bar").show(ctx, |ui| {
                    ui.with_layout(Layout::top_down(Align::Center), |ui| {
                        ui.label(RichText::new("An error has occured :(").size(32.0));
                    });
                });

                egui::CentralPanel::default().show(ctx, |ui| {
                    center_vertical(ui,|ui| { 
                        ui.label(RichText::new(reason).size(32.0));
                    });
                });
            },
            UiState::KeyRequired(reason) => {
                let mut is_valid = true;
                let hash = sha2::Sha256::digest(&self.key_input.as_bytes());
                for i in 0..32 {
                    if hash[i] != GAME_KEY_HASH[i] {
                        is_valid = false;
                    }
                }

                egui::TopBottomPanel::top("key_prompt_title_bar").show(ctx, |ui| {
                    ui.with_layout(Layout::top_down(Align::Center), |ui| {
                        ui.label(RichText::new("Decryption key required").size(32.0));
                        ui.label(RichText::new(format!("Reason: {reason}")).size(24.0));
                    });
                });

                egui::CentralPanel::default().show(ctx, |ui| {
                    top_horizontal(ui, |ui| {
                        ui.horizontal(|ui| {
                            ui.label("Decryption key: ");
                            ui.text_edit_singleline(&mut self.key_input);
                            if is_valid {
                                ui.label("Valid");
                            } else {
                                ui.label(RichText::new("Invalid").color(ui.visuals().error_fg_color));
                            }
                        });
                    });
                });

                egui::TopBottomPanel::bottom("key_prompt_button_bar").show(ctx, |ui| {
                    if !is_valid {
                        ui.disable();
                    }
                    
                    ui.with_layout(Layout::top_down(Align::Max), |ui| {
                        if ui.button("Continue").clicked() {
                            self.sender.send(UiEvent::SetKey(self.key_input.as_bytes().to_vec())).expect("Failed to send");
                        }
                    })
                });
            },
            UiState::PickGame(steam, others) => {
                let invalid_path_modal = Modal::new(ctx, "invalid_path_modal");
                invalid_path_modal.show(|ui| {
                    invalid_path_modal.title(ui, "Invalid path");
                    invalid_path_modal.frame(ui, |ui| {
                        invalid_path_modal.body(ui, "Please pick a correct installation of OMORI.");
                    });
                    invalid_path_modal.buttons(ui, |ui| {
                        invalid_path_modal.button(ui, "OK");
                    });
                });
                egui::TopBottomPanel::top("key_prompt_title_bar").show(ctx, |ui| {
                    ui.with_layout(Layout::top_down(Align::Center), |ui| {
                        ui.label(RichText::new("Select your base game").size(32.0));
                    });
                });

                egui::CentralPanel::default().show(ctx, |ui| {
                    ui.columns(2, |columns| {
                        columns[0].vertical(|ui| {
                            ui.label(RichText::new("Steam Installation").size(24.0));
                            match steam {
                                Ok(path) => {
                                    ui.label(RichText::new(format!("Found game at:\n{}", path.display())).italics());
                                    ui.separator();
                                    if ui.button(RichText::new("Use Steam version").size(16.0)).clicked() {
                                        self.sender.send(UiEvent::UseSteamPath).expect("Failed to send");
                                    }
                                },
                                Err(e) => {
                                    ui.label(RichText::new(format!("Failed to find game:\n{}", e)).color(ui.visuals().error_fg_color));
                                }
                            }
                        });
                        
                        columns[1].vertical(|ui| {
                            ui.label(RichText::new("Custom").size(24.0));
                            if ui.button(RichText::new("Pick game location").size(16.0)).clicked() {
                                match rfd::FileDialog::new().pick_folder() {
                                    Some(path) => {
                                        if omori_locator::validate_omori_installation(&path) {
                                            self.sender.send(UiEvent::UsePath(path)).expect("Failed to send");
                                        } else {
                                            invalid_path_modal.open();
                                        }
                                    },
                                    None => {}
                                }
                            }
                            ui.separator();

                            if others.len() > 0 {
                                ui.label("History of game locations");
  
                                TableBuilder::new(ui)
                                    .striped(true)
                                    .column(Column::remainder())
                                    .column(Column::auto())
                                    .body(|body| {
                                        body.rows(20.0, others.len(), |mut row| {
                                            let item = &others[row.index()];
                                            let dt = 
                                                (item.1 as i64) -
                                                (SystemTime::now().duration_since(SystemTime::UNIX_EPOCH).expect("Failed to get the time").as_secs() as i64);
                                                
                                            let dt = chrono::Duration::from(chrono::TimeDelta::seconds(dt as i64));

                                            row.col(|ui| {
                                                ui.label(format!("{}", item.0.display()));
                                            });
                                            
                                            row.col(|ui| {
                                                ui.label(format!("{}", chrono_humanize::HumanTime::from(dt)));
                                            });
                                        });
                                    });

                            } else {
                                ui.label("Once you use a custom location, it will be remembered here.");
                            }
                        });
                    });
                });
            }
        }
    }
}